from typing import Optional

from anaconda_auth import login as anaconda_login
from anaconda_auth import logout as anaconda_logout
from anaconda_auth.exceptions import TokenNotFoundError
from anaconda_auth.handlers import shutdown_all_servers
from anaconda_auth.token import TokenInfo


class AuthService:
    def login(self):
        """
        Attempts to retrieve an API key for authentication.

        This method first tries to obtain an API key from the current session.
        If this fails, it forcibly shuts down all servers, retries the login process,
        and then attempts to retrieve the API key again. If both attempts fail,
        it returns None to indicate that a valid API key cannot be obtained.

        Returns:
            str: The retrieved API key if successful, or None if authentication fails.
        """
        if token_info := self.get_token():
            return token_info.api_key

        try:
            shutdown_all_servers()
            # Logs in on ananconda.com, not anaconda.cloud
            anaconda_login()

            if token_info := self.get_token():
                return token_info.api_key
        except (TokenNotFoundError, Exception):
            return None

    def logout(self) -> None:
        """
        Logs out of anaconda.com and fallback on anaconda.cloud.

        NOTE: we don't use anaconda_auth.logout because this one fails silently.

        :return: None
        """
        token_info = self.get_token()
        if token_info and hasattr(token_info, "api_key"):
            anaconda_logout()
        return None

    def get_token(self) -> Optional[str]:
        """
        Retrieves the API key for anaconda.com and fallback on anaconda.cloud.

        Returns:
            str: The loaded API key or None.
        """
        try:
            return TokenInfo.load()
        except TokenNotFoundError:
            return None


auth_service = AuthService()
